#include <d3dx8.h>
#include <stdlib.h>
#include <time.h>
#include "QuatFrac.h"

// Max number of iterations
#define MAXITERS 30
// Maximum value (considered to be infinite)
#define MAXVALUE 4.f
#define MAXVALUE2 ( MAXVALUE * MAXVALUE )

// Vector space to span
#define VSPACE_MINX  -50.f
#define VSPACE_MAXX  50.f
#define VSPACE_SIZEX ( VSPACE_MAXX - VSPACE_MINX )

#define VSPACE_MINY  -50.f
#define VSPACE_MAXY  50.f
#define VSPACE_SIZEY ( VSPACE_MAXY - VSPACE_MINY )

#define VSPACE_MINZ  -50.f
#define VSPACE_MAXZ  50.f
#define VSPACE_SIZEZ ( VSPACE_MAXZ - VSPACE_MINZ )

// Quaternion space to span
#define QSPACE_MINX  -2.f
#define QSPACE_MAXX  2.f
#define QSPACE_SIZEX ( QSPACE_MAXX - QSPACE_MINX )

#define QSPACE_MINY  -2.f
#define QSPACE_MAXY  2.f
#define QSPACE_SIZEY ( QSPACE_MAXY - QSPACE_MINY )

#define QSPACE_MINZ  -2.f
#define QSPACE_MAXZ  2.f
#define QSPACE_SIZEZ ( QSPACE_MAXZ - QSPACE_MINZ )

// Light colors
static const D3DXVECTOR3 globLightCols[] = {
//	D3DXVECTOR3( 1.0f, 1.0f, 1.0f ),
	D3DXVECTOR3( 0.7f, 0.5f, 1.0f ),
	D3DXVECTOR3( 0.5f, 1.0f, 0.7f ),
	D3DXVECTOR3( 1.0f, 0.7f, 0.5f ),
	D3DXVECTOR3( 1.0f, 1.0f, 0.0f ),
	D3DXVECTOR3( 0.0f, 1.0f, 1.0f ),
	D3DXVECTOR3( 1.0f, 0.0f, 1.0f ),
};
static const int globLightColNum = sizeof( globLightCols ) / sizeof( D3DXVECTOR3 );

// Converts an index point coordinate to space coords
inline void idxcoordToSpace( D3DXVECTOR3 &dest, int x, int y, int z )
{
	dest.x = VSPACE_MINX + VSPACE_SIZEX * ( ( float ) ( x - ISPACE_MINX ) / ( float ) ISPACE_SIZEX );
	dest.y = VSPACE_MINY + VSPACE_SIZEY * ( ( float ) ( y - ISPACE_MINY ) / ( float ) ISPACE_SIZEY );
	dest.z = VSPACE_MINZ + VSPACE_SIZEZ * ( ( float ) ( z - ISPACE_MINZ ) / ( float ) ISPACE_SIZEZ );
}
inline void idxcoordToSpace( D3DXQUATERNION &dest, int x, int y, int z )
{
	dest.x = QSPACE_MINY + QSPACE_SIZEY * ( ( float ) ( y - ISPACE_MINY ) / ( float ) ISPACE_SIZEY );
	dest.y = QSPACE_MINZ + QSPACE_SIZEZ * ( ( float ) ( z - ISPACE_MINZ ) / ( float ) ISPACE_SIZEZ );
	dest.z = 0.f;
	dest.w = QSPACE_MINX + QSPACE_SIZEX * ( ( float ) ( x - ISPACE_MINX ) / ( float ) ISPACE_SIZEX );
}

// Tests if a given index space point is in valid range
inline bool isValid( int x, int y, int z )
{
	return !( x < ISPACE_MINX || x > ISPACE_MAXX ||
			  y < ISPACE_MINY || y > ISPACE_MAXY ||
			  z < ISPACE_MINZ || z > ISPACE_MAXZ );
}

////

QuaternionFractal::QuaternionFractal()
{
	myVertices.reserve( 100000 );

	srand( time( 0 ) );
	initFractal();

	myLightDir = D3DXVECTOR3( 1, 1, 0 );
	D3DXVec3Normalize( &myLightDir, &myLightDir );
}

// Initializes the fractal calculator to calculate a random fractal
void QuaternionFractal::initFractal()
{
	float q[ 4 ];
	int i;

//	initFractal( D3DXQUATERNION( 0.2f, 0, 0, -1 ) );

	// Pick random values for the start value
	for ( i = 0; i < 4; ++i )
	{
		q[ i ] = ( float ) ( rand() % 100000 ) / 100000.f;
		if ( ( rand() % 100 ) < 50 ) q[ i ] = -q[ i ];
	}

	// Subject the quaternion for random mutations
	int mutations = rand() & ( ( 1 << 2 ) - 1 );

	if ( mutations & 1 )
	{ // Set x values to the same / negative
		float val = q[ rand() % 4 ];
		int j = 1 + ( rand() % 4 );

		for ( i = 0; i < j; ++i )
		{
			q[ rand() % 4 ] = ( rand() % 10 ) < 5 ? val : -val;
		}
	}

	if ( mutations & 2 )
	{ // Set x values to zero
		int j = 1 + ( rand() % 2 );

		for ( i = 0; i < j; ++i )
			q[ rand() % 4 ] = 0.f;
	}

	initFractal( D3DXQUATERNION( q[ 0 ], q[ 1 ], q[ 2 ], q[ 3 ] ) );
}

// Initializes the fractal calculator to calculate a given fractal
void QuaternionFractal::initFractal( D3DXQUATERNION quat )
{
	myQuatC = quat;

	myVertices.clear();
	myTestedVertices.reset();
	myFractalVertices.reset();

	while ( !myVertexProcessQueue.empty() )
		myVertexProcessQueue.pop();

	// Pick a new color
	myLightCol = globLightCols[ rand() % globLightColNum ];
}

// Computes the fractal n points more
void QuaternionFractal::compute( int points )
{
	BoolMatrix isFilled;

	// Init
	int x, y, z;

	// Check the points
	while ( points-- > 0 )
	{
		// Any vertices to process?
		if ( myVertexProcessQueue.size() > 0 )
		{
			int idx = myVertexProcessQueue.top();
			myVertexProcessQueue.pop();

			// Get the coords
			indexToCoord( idx, x, y, z );
		}
		else
		{ // Select an untested vertex
			int left = 20;
			while ( left-- > 0 )
			{
				x = ISPACE_MINX + ( rand() % ISPACE_SIZEX );
				y = ISPACE_MINY + ( rand() % ISPACE_SIZEY );
				z = ISPACE_MINZ + ( rand() % ISPACE_SIZEZ );

				if ( myTestedVertices.test( coordToIndex( x, y, z ) ) == false )
					break;
			}
		}

		// Does this belong to a fractal?
		if ( !testPoint( x, y, z ) )
			continue;

		// Process the surrounding points (->queue)
		bool emptyFound = false;
		for ( int tx = -1; tx <= 1; ++tx )
		{
			for ( int ty = -1; ty <= 1; ++ty )
			{
				for ( int tz = -1; tz <= 1; ++tz )
				{
					isFilled[ tx + 1 ][ ty + 1 ][ tz + 1 ] = false;

					// Not the already-checked point
					if ( tx == 0 && ty == 0 && tz == 0 )
						continue;
					// Valid coord?
					if ( !isValid( x + tx, y + ty, z + tz ) )
					{
						emptyFound = true;
						continue;
					}

					// Already tested
					int idx = coordToIndex( x + tx, y + ty, z + tz );
					if ( myTestedVertices.test( idx ) )
					{
						if ( !myFractalVertices.test( idx ) )
							emptyFound = true;
						else
							isFilled[ tx + 1 ][ ty + 1 ][ tz + 1 ] = true;
						continue;
					}

					// Test it
					if ( testPoint( x + tx, y + ty, z + tz ) )
					{
						isFilled[ tx + 1 ][ ty + 1 ][ tz + 1 ] = true;
						myVertexProcessQueue.push( idx );
					}
				}
			}
		}

		// If we are not in the middle of a fractal (ie. this is a boundary vertex), add it
		if ( emptyFound )
		{
			FractalVertex v;

			// Create
			idxcoordToSpace( v.v, x, y, z );
			float shade = calcShade( isFilled );

			int base = 100;
			float rest = shade * ( 255 - base );
			D3DXVECTOR3 vec = v.v;

			int r = base + ( int ) ( rest * fabs( sin( D3DXToRadian( vec.x * vec.y + vec.z ) ) ) );
			int g = base + ( int ) ( rest * fabs( sin( D3DXToRadian( vec.x + vec.y * vec.z ) ) ) );
			int b = base + ( int ) ( rest * fabs( sin( D3DXToRadian( vec.x + vec.y + vec.z ) ) ) );
			r = ( int ) ( r * myLightCol.x );
			g = ( int ) ( g * myLightCol.y );
			b = ( int ) ( b * myLightCol.z );
			v.color = D3DCOLOR_ARGB( 255, r, g, b );

/*			int gray = 100;
			gray = gray + ( int ) ( ( 255 - gray ) * shade );
			v.color = D3DCOLOR_ARGB( 255, gray, gray, 0 );*/

			// Add
			myVertices.push_back( v );
		}
	}
}

// Tests if the given index point is in the fractal or not
bool QuaternionFractal::testPoint( int x, int y, int z )
{
	// Out of scope?
	if ( !isValid( x, y, z ) )
		return false;

	// Already tested?
	int idx = coordToIndex( x, y, z );
	if ( myTestedVertices.test( idx ) )
		return myFractalVertices.test( idx );

	// Create the quaternion for the point
	static D3DXQUATERNION quat;
	idxcoordToSpace( quat, x, y, z );

	// Iterate the point
	int left = MAXITERS;
	bool isFrac = true;
	while ( left-- > 0 )
	{
		// Iterate z(n+1) = z(n) + c
//		quat = ( quat * quat * quat ) + myQuatC;
		( quat *= quat ) += myQuatC;

		// Too long? (test every now and then)
		float len = D3DXQuaternionLengthSq( &quat );
		if ( len >= MAXVALUE2 )
		{
			isFrac = false;
			break;
		}
	}

	// Return the result
	myTestedVertices.set( idx, true );
	myFractalVertices.set( idx, isFrac );
	return isFrac;
}

// Calculates a shading value (0-1) when given an adjancency matrix [x][y][z]
float QuaternionFractal::calcShade( BoolMatrix &matrix )
{
	D3DXVECTOR3 normal = D3DXVECTOR3( 0, 0, 0 );

	// Flatten the map
	for ( int x = 0; x < 3; ++x )
		for ( int y = 0; y < 3; ++y )
			for ( int z = 0; z < 3; ++z )
			{
				float nx = -( x - 1 );
				float ny = -( y - 1 );
				float nz = -( z - 1 );

				if ( !matrix[ x ][ y ][ z ] )
					continue;

				normal.x += nx;
				normal.y += ny;
				normal.z += nz;
			}

	// Get the normalized vector
	D3DXVec3Normalize( &normal, &normal );

	// Get the shade against a light (-1 to 1)
	float shade = D3DXVec3Dot( &normal, &myLightDir );

	// Make the light a little excess
/*	shade *= 2.f;
	if ( shade < -1 )
		shade = -1;
	if ( shade > 1 )
		shade = 1;*/

	// Return the shading multiplier (0 to 1)
//	return ( 1.f + shade ) / 2.f;
	return ( shade >= 0 ) ? shade : -shade;
}
